/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file contours.cpp
/// @brief Example program for QVision library. Gets contours from thresholded copy of the input image
///
/// This program will receive a path or url to a video file, and then will show hot points or corners of each frame in a
/// window. It detects corners where the minimal eigenvalue of the first derivatives matrix for a pixel is greater than a
/// Threshold value.
/// To use the program, simply compile, and execute the binary:
///
///	./contours --Video=<path or url to video>
///
/// @author PARP Research Group. University of Murcia, Spain.
///

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <QDebug>

#include <qvcore/qvapplication.h>
#include <qvmplayercamera/qvmplayercamera.h>
#include <qvgui/qvisioninterface.h>
#include <qvdta/qvpolyline.h>
#include <qvdta/qvcontour.h>

class MyWorker: public QVWorker
	{
	public:
		MyWorker(QString name): QVWorker(name)
			{
			addProperty<int>("Threshold", inputFlag, 128, "Umbral para que un punto cuente como distinto de cero", 0, 256);
			addProperty< QVImage<uChar,1> >("Input image", inputFlag|outputFlag);
			addProperty< QVImage<uChar,3> >("Contours", outputFlag);
			}

		void worker()
			{
			////////////////////
			// Read input parameters
			QVImage<uChar> image = getPropertyValue< QVImage<uChar,1> >("Input image");
			int threshold = getPropertyValue< int >("Threshold");
			uInt rows = image.getRows(), cols = image.getCols();
			timeFlag("Read input parameters");
		
			///////////////////
			// Read contours from image
			QList<qvdta::QVPolyline> contours = qvdta::getBorderContoursThreshold(image, threshold);
			timeFlag("Read contours from image");

			///////////////////
			// Separate inside and outside contours
			QList<qvdta::QVPolyline> directos, reversos;
			for (int i = 0; i< contours.size(); i++)
				if(contours.at(i).direction)
					directos.append(contours.at(i));
				else
					reversos.append(contours.at(i));
			timeFlag("Separate inside and outside contours");
		
			//////////////////
			// Draw contours in images
			QVImage<uChar> outsideContoursImage(cols, rows), insideContoursImage(cols, rows);
			qvipp::Set(outsideContoursImage,0);
			qvipp::Set(insideContoursImage,0);

			draw(outsideContoursImage, directos, (uChar) 255);
			draw(insideContoursImage, reversos, (uChar) 255);
			timeFlag("Draw contours in images");

			//////////////////////
			// Compose  and show result image
			QVImage<uChar,3> thresholded(cols, rows);
			qvipp::Set(thresholded, 0,0,0);

			for(uInt col = 0; col < cols; col++)
				for(uInt row = 0; row < rows; row++)
					{
					if (image(col, row) >= threshold)
						{
						thresholded(col, row,0) = 64;
						thresholded(col, row,1) = 64;
						thresholded(col, row,2) = 64;
						}

					if (outsideContoursImage(col, row) == 255) thresholded(col, row,0) = 255;
					if (insideContoursImage(col, row) == 255) thresholded(col, row,1) = 255;
					}
			
			setPropertyValue< QVImage<uChar,3> >("Contours", thresholded);
			timeFlag("Compose and show result image");
			}
	};



int main(int argc, char *argv[])
	{
	QVApplication app(argc, argv,
		
		"Example program for QVision library. Applies corner detection over an input video."

		);

	QVMPlayerCamera camera("Video");
	MyWorker worker("Corners Worker");
	QVisionInterface interface;

	camera.link(&worker,"Input image");
	interface.setVisible(&worker,"Input image");
	interface.setVisible(&worker,"Contours");

	return app.exec();
	}
