/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file rotoscoper.cpp
/// @brief Example program for QVision library. Composes component tree image filtering and canny operator for making animation like images from
/// real images.
///
/// This program will receive a path or url to a video file, and then will show in a window the resulting rostoscoped video.
/// To use the program, simply compile, and execute the binary:
///
///	./rotoscoper --Video=<path or url to video>
///
/// @author PARP Research Group. University of Murcia, Spain.
///

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <QDebug>

#include <qvcore/qvapplication.h>
#include <qvmplayercamera/qvmplayercamera.h>
#include <qvgui/qvisioninterface.h>

#include <qvdta/qvdta.h>
#include <qvdta/qvcomponenttree.h>

class ComponentTreeWorker: public QVWorker
	{
	public:
		ComponentTreeWorker(QString name): QVWorker(name)
			{
			addProperty<int>("Maximal area to prune", inputFlag, 50,
				"Maximal size of the areas to be pruned in the image", 5, 320*240);
			addProperty< QVImage<uChar,1> >("Input image", inputFlag|outputFlag);
			addProperty< QVImage<uChar,1> >("Pruned component tree image", outputFlag);
			}

		void worker()
			{
			//////////////////
			// Read parameters
			QVImage<uChar> image = getPropertyValue< QVImage<uChar,1> >("Input image");
			const uInt	rows = image.getRows(), cols = image.getCols();
			const uInt minAreaThreshold = getPropertyValue<int>("Maximal area to prune");
			timeFlag("Read parameters");
		
			//////////////////////////////////////////////////
			// Prune low areas from image, with component tree
			qvdta::QVComponentTree componentTreeLow(image,true);
			timeFlag("Call to getComponentTree for low areas");

			qvdta::FilterComponentTreeSmallRegions(image, componentTreeLow, minAreaThreshold);
			timeFlag("Prune low areas from image");

			///////////////////////////////////////////////////
			// Prune high areas from image, with component tree
			qvdta::QVComponentTree componentTreeHigh(image);
			timeFlag("Call to getComponentTree for high areas");

			qvdta::FilterComponentTreeSmallRegions(image, componentTreeHigh, minAreaThreshold);
			timeFlag("Prune high areas from image");
		
			/////////////////
			// Publish resulting images
			setPropertyValue< QVImage<uChar,1> >("Pruned component tree image", image);
			timeFlag("Publish resulting images");
			}
	};

class CannyWorker: public QVWorker
	{
	public:
		CannyWorker(QString name): QVWorker(name)
			{
			addProperty<double>("Threshold high", inputFlag,	150,	"High threshold for Canny operator", 50, 1000);
			addProperty<double>("Threshold low", inputFlag,		50,	"Low threshold for Canny operator", 10, 500);
			addProperty< QVImage<uChar,1> >("Canny image", outputFlag);
			addProperty< QVImage<uChar,1> >("Input image", inputFlag|outputFlag);
			}

		void worker()
			{
			//////////////////
			// Read input parameters
			QVImage<uChar,1> image = getPropertyValue< QVImage<uChar,1> >("Input image");
			uInt cols = image.getCols(), rows = image.getRows();
			QVImage<sFloat>	imageFloat(cols, rows), dX(cols, rows), dY(cols, rows), dXNeg(cols, rows);
			QVImage<uChar>	canny(cols, rows), buffer;
		
			//////////////////
			// Convert image from uChar to sShort
			qvipp::Convert(image, imageFloat);
			timeFlag("Convert image from uChar to sShort");
		
			//////////////////
			// Obtain horizontal and vertical gradients from image
			qvipp::FilterSobelHorizMask(imageFloat,dY,3);
			qvipp::FilterSobelVertMask(imageFloat,dX,3);
			qvipp::MulC(dX, dXNeg, -1);
			timeFlag("Obtain horizontal and vertical gradients from image");
		
			/////////////////
			// Apply Canny operator
			qvipp::CannyGetSize(canny, buffer);
			qvipp::Canny(dXNeg, dY, canny, buffer, getPropertyValue<double>("Threshold low"), getPropertyValue<double>("Threshold high"));
			timeFlag("Apply Canny operator");
		
			/////////////////
			// Publish resulting images
			setPropertyValue< QVImage<uChar,1> >("Canny image",canny);
			timeFlag("Publish resulting images");
			}
	};


class RotoscoperWorker: public QVWorker
	{
	public:
		RotoscoperWorker(QString name): QVWorker(name)
			{
			addProperty< QVImage<uChar,1> >("Borders image", inputFlag|outputFlag);
			addProperty< QVImage<uChar,1> >("Flat colors image", inputFlag|outputFlag);
			addProperty< QVImage<uChar,1> >("Output image", outputFlag);
			}

		void worker()
			{
			//////////////////
			// Read parameters
			QVImage<uChar> bordersImage = getPropertyValue< QVImage<uChar,1> >("Borders image");
			QVImage<uChar> flatColorsImage = getPropertyValue< QVImage<uChar,1> >("Flat colors image");
			uInt	rows = bordersImage.getRows(), cols = bordersImage.getCols();
			timeFlag("Read parameters");
		
			//////////////////
			// Paint contours		
			for(uInt col = 0; col < cols; col++)
				for (uInt row = 0; row < rows; row++)
					if (bordersImage(col, row))
						flatColorsImage(col, row) = 0;

			/////////////////
			// Publish resulting images
			setPropertyValue< QVImage<uChar,1> >("Output image", flatColorsImage);
			timeFlag("Publish resulting images");
			}
	};

int main(int argc, char *argv[])
	{
	QVApplication app(argc, argv,
		
		"Example program for QVision library. Gets component tree from images."

		);
	
	ComponentTreeWorker componentTreeWorker("Component Tree");
	CannyWorker cannyWorker("Canny");
	RotoscoperWorker rotoscoperWorker("Rotoscoper");

	componentTreeWorker.linkProperty("Pruned component tree image", &cannyWorker, "Input image", QVWorker::SynchronousLink);
	cannyWorker.linkProperty("Canny image", &rotoscoperWorker, "Borders image", QVWorker::SynchronousLink);
	cannyWorker.linkProperty("Input image", &rotoscoperWorker, "Flat colors image", QVWorker::SynchronousLink);

	QVMPlayerCamera camera("Video");
	camera.link(&componentTreeWorker,"Input image");

	QVisionInterface interface;
	interface.setVisible(&rotoscoperWorker,"Output image");

	return app.exec();
	}


	
