/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvcamera.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVCAMERA_H
#define QVCAMERA_H

#include <QObject>
#include <QString>
#include <QDebug>
#include <qvcore/qvimage.h>
#include <qvcore/qvpropertyholder.h>

/// @class QVCamera qvcore/qvcamera.h
/// @brief (Virtual) parent class for every camera.
///
/// Declares an interface for every camera. By now, QVMplayerCamera is the
/// only real camera that is implemented (inheriting from this class), but
/// in the future other cameras (direct interface to IEEE 1394, usb cameras,
/// etc.) could be implemented using the same interface.
///
/// Note.- This class is not intended to be used by the final user of the
///        library; It will be useful only for those wishing to implement new
///        camera types.
class QVCamera : public QObject, public QVPropertyHolder
{
/// \cond  Not documented by now.
	Q_OBJECT
public:
	QVCamera(QString name = QString());
	~QVCamera();

	// Camera status:
	typedef enum {Closed, Running, RunningOneStep, Paused /*, EndOfFile */} TCameraStatus;

	/// @brief Options for a seek through the file.
	typedef enum { SeekCurrent = 0, SeekPercentage = 1, SeekAbsolute = 2 } TSeekType;

	virtual bool openCam() = 0;

	//virtual bool isEndOfFile() const { return status == EndOfFile; };

	// Getter functions:
	virtual unsigned int getRows() const = 0;
	virtual unsigned int getCols() const = 0;
	virtual double getFPS() const = 0;
	virtual double getSpeed() const = 0;
	virtual unsigned int getFramesGrabbed() const = 0;
	virtual unsigned int getFramesRead() const = 0;
	virtual const QString getUrlBase() const = 0;
	virtual double getTimeLength() const = 0;
	virtual double getTimePos() const = 0;
	virtual bool isLiveCamera() const = 0;
	virtual TCameraStatus getCameraStatus() const { return status; };
	virtual bool isClosed() const { return status == Closed; };
	virtual bool isRunning() const { return status == Running; };
	virtual bool isRunningOneStep() const { return status == RunningOneStep; };
	virtual bool isPaused() const { return status == Paused; };


public slots:
	virtual bool grab() = 0;
	virtual void pauseCam() = 0;
	virtual void unpauseCam() = 0;
	virtual void nextFrameCam() = 0;
	virtual void setSpeedCam(double d) = 0;
	virtual void seekCam(TSeekType type, double pos) = 0;
	virtual void closeCam() = 0;

signals:
	void newGrab();
	void newRead();
	void camClosed();
	void camOpened();
	void statusChange(QVCamera::TCameraStatus);

protected:
	TCameraStatus status;
	void setStatus(TCameraStatus status) 
		{ this->status = status; emit statusChange(status); }

	void setImageBuffer(QVGenericImage * image)
		{
		if (this->image != NULL)
		delete image;
		this->image = image;
		}

private:
	QVGenericImage * image;

/// \endcond  Not documented by now.
};
#endif

