/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvpropertyholder.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include "qvpropertyholder.h"

QVPropertyHolder::QVPropertyHolder(const QString name):
	name(name), errorString(), variants(), safelyCopiedVariants(), minimum(),
	maximum(), _info(), io_flags(), link_flags(), insertion_order(),
	inputLinks(), outputLinks()
	{
	if(qvApp == NULL)
		{
		QString str = "QVPropertyHolder::QVPropertyHolder(): holder " + name +
					  ": property holders cannot be created before the " +
					  "QVApplication instance. Aborting now.";
		std::cerr << qPrintable(str) << std::endl;
		exit(1);
		}
	else
		qvApp->registerQVPropertyHolder(this);
	}

QVPropertyHolder::~QVPropertyHolder()
{
	if(qvApp != NULL)
		qvApp->deregisterQVPropertyHolder(this);
	// Not needed now: unlink();
}

void QVPropertyHolder::setName(const QString name)
	{ this->name = name; }

const QString QVPropertyHolder::getName() const
	{ return this->name; }

QList<QString> QVPropertyHolder::getPropertyList() const
	{ return variants.keys(); }

bool QVPropertyHolder::containsProperty(const QString name) const
	{ return variants.contains(name); }

QVariant::Type QVPropertyHolder::getPropertyType(const QString name, bool *ok) const
	{
	if(not checkExists(name,"QVPropertyHolder::getPropertyType()"))
		{
		if(ok != NULL) *ok = FALSE;
		return QVariant::Invalid;
		}
	if(ok != NULL) *ok = TRUE;
	QVariant variant = variants.value(name);
	return variant.type();
	}

bool QVPropertyHolder::removeProperty(const QString name)
	{
	if(not checkExists(name,"QVPropertyHolder::removeProperty()"))
		return FALSE;
	this->variants.remove(name);
	this->safelyCopiedVariants.remove(name);
	this->minimum.remove(name);
	this->maximum.remove(name);
	this->_info.remove(name);
	this->io_flags.remove(name);
	int i = this->insertion_order.indexOf(name);
	this->insertion_order.removeAt(i);
	return TRUE;
	}

bool QVPropertyHolder::setPropertyRange(const QString name, const double & minimum, const double & maximum)
	{
	if(not checkExists(name,"QVPropertyHolder::setPropertyRange()"))
		return FALSE;
	if(minimum <= getPropertyValue<double>(name) and
		maximum >= getPropertyValue<double>(name))
		{
			this->minimum[name] = QVariant::fromValue(minimum);
			this->maximum[name] = QVariant::fromValue(maximum);
			return TRUE;
		} else {
			QString str =  "QVPropertyHolder::setPropertyRange(): property " +
						   name + " in holder " + getName() + " has value " +
						   QString("%1").arg(getPropertyValue<double>(name)) +
						   ", which is not valid for the range [" +
						   QString("%1,%2").arg(minimum).arg(maximum) + "]." ;
			setLastError(str);
			if(qvApp->isRunning()) {
				std::cerr << qPrintable("Warning: " + str + "\n");
			} // Otherwise, qApp will show the error and won't start the program.
			return FALSE;
		}
	}

bool QVPropertyHolder::setPropertyRange(const char *name, int & minimum, int & maximum)
	{ return setPropertyRange(QString(name), static_cast<double>(minimum), static_cast<double>(maximum)); }

bool QVPropertyHolder::setPropertyRange(QString name, int & minimum, int & maximum)
	{ return setPropertyRange(name, static_cast<double>(minimum), static_cast<double>(maximum)); }

bool QVPropertyHolder::hasRange(const QString name)
	{ return maximum.contains(name) and minimum.contains(name); }

bool QVPropertyHolder::isInput(const QString name)
	{ return (io_flags[name] & inputFlag);};

bool QVPropertyHolder::isOutput(const QString name)
	{ return (io_flags[name] & outputFlag);};

bool QVPropertyHolder::isLinkedInput(const QString name)
	{ return (link_flags[name] & linkedInputFlag);};

bool QVPropertyHolder::isLinkedOutput(const QString name)
	{ return (link_flags[name] & linkedOutputFlag);};

QVariant QVPropertyHolder::getPropertyQVariantValue(const QString name, bool *ok) const
	{
	if (not checkExists(name,"QVPropertyHolder::getPropertyQVariantValue()"))
		if(ok != NULL) *ok = FALSE;
	else
		if(ok != NULL) *ok = TRUE;
	return variants[name];
	}

QString QVPropertyHolder::getPropertyInfo(const QString name, bool *ok) const
	{
	if(not checkExists(name,"QVPropertyHolder::getPropertyInfo()"))
		if(ok != NULL) *ok = FALSE;
	else
		if(ok != NULL) *ok = TRUE;
	return this->_info[name];
	}

QString QVPropertyHolder::getLastError() const
{
	return errorString;
}

const QString QVPropertyHolder::infoInputProperties() const
	{
	QString info = QString("Input parameters for ") + getName() + QString(":\n");
	bool emptyInfo=TRUE;

	QStringList props;
	QListIterator<QString> i(insertion_order);
	while (i.hasNext())
		{
		QString property = i.next();
		if(not (io_flags[property] & inputFlag) or (link_flags[property] & linkedInputFlag))
			continue;
		emptyInfo=FALSE;
		info += "  --" + property + "=";
		QVariant::Type type = getPropertyType(property);
		switch(type)
			{
			case QVariant::String:
				{
				READ_PROPERTY(this, QString, property, stringValue);
				info += "[text] ";
				if(stringValue != QString())
					info += "(def. " + stringValue + ") ";
				}
				break;
			case QVariant::Double:
				{
				READ_PROPERTY(this,double,property,doubleValue);
				if(maximum.contains(property) and minimum.contains(property))
					{
					READ_PROPERTY_MAX(this,double,property,doubleMaximum);
					READ_PROPERTY_MIN(this,double,property,doubleMinimum);
					info += "[" + QString().setNum(doubleMinimum) + "..."
						+ QString().setNum(doubleMaximum) + "] ";
					}
				else
					info += "[double] ";
				info += "(def. "+ QString().setNum(doubleValue) + ") ";
				}
				break;
			case QVariant::Int:
				{
				READ_PROPERTY(this,int,property,intValue);
				if(maximum.contains(property) and minimum.contains(property))
					{
					READ_PROPERTY_MAX(this,int,property,intMaximum);
					READ_PROPERTY_MIN(this,int,property,intMinimum);
					info += "[" + QString().setNum(intMinimum) + "..."
						+ QString().setNum(intMaximum) + "] ";
					}
				else
					info += "[int] ";
					info += "(def. "+ QString().setNum(intValue) + ") ";
				}
				break;
			case QVariant::Bool:
				{
				READ_PROPERTY(this, bool, property, boolValue);
				info += "[true,false]"
					+ (boolValue ? QString(" (def. true) "):QString("(def. false) "));
				}
				break;
			default:
				info += QString("[%1] ").arg(type);
				break;
			}
		QString propertyInfo = getPropertyInfo(property);
		if (propertyInfo != QString())
			propertyInfo = " " + propertyInfo + ". ";
		info += propertyInfo.rightJustified(80-info.split('\n').last().length(),'.') + "\n";
		}
	if(emptyInfo)
		return QString(""); //info += "  None.\n";

	return info;
	}

bool QVPropertyHolder::correctRange(const QString name, const double & value) const
	{
	if(not maximum.contains(name) and not minimum.contains(name))
		return TRUE;
	double maximum = getPropertyMaximum<double>(name);
	double minimum = getPropertyMinimum<double>(name);
	if(minimum <= value and maximum >= value)
		return TRUE;
	else
		{
		QString str =  "QVPropertyHolder::setPropertyValue(): value " +
					   QString("%1").arg(value) + " for property " +
					   name + " in holder " + getName() +
					   "is not valid for the range [" +
					   QString("%1,%2").arg(minimum).arg(maximum) + 
					   "] stablished for it." ;
		setLastError(str);
		if(qvApp->isRunning())
			{
			std::cerr << qPrintable("Warning: " + str + "\n");
			} // Otherwise, qApp will show the error and won't start the program.
		return FALSE;
		}
	}

bool QVPropertyHolder::correctRange(const char *name, const int & value) const
	{ return correctRange(QString(name),static_cast<double>(value)); }

bool QVPropertyHolder::correctRange(QString name, const int & value) const
	{ return correctRange(name,static_cast<double>(value)); }

bool QVPropertyHolder::checkExists(const QString name, const QString methodname) const
	{
	if(not variants.contains(name))
		{
		QString str =  methodname + ": property " + name +
					   " doesn't exists in holder " + getName() + ".";
		setLastError(str);
		if(qvApp->isRunning()) {
			std::cerr << qPrintable("Warning: " + str + "\n");
		} // Otherwise, qApp will show the error and won't start the program.
		return FALSE;
		} else {
		return TRUE;
		}
	}

bool QVPropertyHolder::checkIsNewProperty(const QString name, const QString methodname) const
	{
	if(variants.contains(name))
		{
		QString str =  methodname + "(): property " + name +
					   " already exists in holder " + getName() + ".";
		setLastError(str);
		if(qvApp->isRunning()) {
			std::cerr << qPrintable("Warning: " + str + "\n");
		} // Otherwise, qApp will show the error and won't start the program.
		return FALSE;
		} else {
		return TRUE;
		}
	}

bool QVPropertyHolder::linkProperty(QString prop_orig,QVPropertyHolder *qvp_dest,QString prop_dest,LinkType link_type)
	{
	bool ok1,ok2;
	QString errMsg;
	QVariant::Type t1,t2;
	QVPropertyHolder *qvp_orig=this, *qvp_err=NULL;
	
	t1 = qvp_orig->getPropertyType(prop_orig,&ok1);
	t2 = qvp_dest->getPropertyType(prop_dest,&ok2);
	if(qvApp->isRunning())
		{
		qvp_err = qvp_orig;
		errMsg = QString("QVPropertyHolder::linkProperty(): Property holder %1:"
					"Cannot link properties after launching QVApplication.\n")
					.arg(prop_orig).arg(qvp_orig->getName());
		}
	else if(qvp_orig == qvp_dest)
		{
		errMsg = QString("QVPropertyHolder::linkProperty(): Property holder %1: cannot link a QVPropertyHolder with itself.\n").arg(qvp_orig->getName());
		qvp_err = qvp_orig;
		}
	else if(not ok1)
		{
		errMsg = QString("QVPropertyHolder::linkProperty(): Property %1 does not exist in property holder %2.\n")
				.arg(prop_orig).arg(qvp_orig->getName());
		qvp_err = qvp_orig;
		}
	else if (not ok2)
		{
		errMsg = QString("QVPropertyHolder::linkProperty(): Property %1 does not exist in property holder %2.\n")
				.arg(prop_dest).arg(qvp_dest->getName());
		qvp_err = qvp_dest;
		}
	else if(t1 != t2)
		{
		errMsg = QString("QVPropertyHolder::linkProperty(): Properties %1 and %2 of QVPropertyHolders %3 and %4 respectively are not of the same type.\n").arg(prop_orig).arg(prop_dest).arg(qvp_orig->getName()).arg(qvp_dest->getName());
		qvp_err = qvp_orig;
		}
	else if(not (qvp_orig->io_flags[prop_orig] & outputFlag))
		{
		errMsg = QString("QVPropertyHolder::linkProperty(): Property %1 of property holder %2 is not of Output type, and cannot be linked as such.\n").arg(prop_orig).arg(qvp_orig->getName());
		qvp_err = qvp_orig;
		}
	else if(not (qvp_dest->io_flags[prop_dest] & inputFlag))
		{
		errMsg = QString("QVPropertyHolder::linkProperty(): Property %1 property holder %2 is not of Input type, and cannot be linked as such.\n").arg(prop_dest).arg(qvp_dest->getName());
		qvp_err = qvp_dest;
		}

	if(errMsg != QString())
		{
		qvp_err->setLastError(errMsg);
		if(qvApp->isRunning()) {
			std::cerr << qPrintable("Warning: " + errMsg + "\n");
		} // Otherwise, qApp will show the error and won't start the program.
		return FALSE;
		}
	else
		{
		QVPropertyHolderLink *link = new
			 QVPropertyHolderLink(qvp_orig,prop_orig,qvp_dest,prop_dest,link_type);
				qvp_orig->outputLinks[prop_orig].push_back(link);
				qvp_dest->inputLinks[prop_dest] = link;
		qvp_dest->link_flags[prop_dest] |= linkedInputFlag;
		qvp_orig->link_flags[prop_orig] |= linkedOutputFlag;
		return TRUE;
		}
	}

void QVPropertyHolder::unlink()
	{
		QMapIterator<QString, QVPropertyHolderLink*> i_in(inputLinks);
		while (i_in.hasNext()) {
			i_in.next();
			QVPropertyHolderLink *link = i_in.value();
			link->markedForDeletion = TRUE;
			// Protect against a possible pending acquire() from our input
			// in other holders:
			link->SyncSemaphoreIn.release();
		}
	
		QMapIterator<QString, QList<QVPropertyHolderLink*> >i_out(outputLinks);
		while (i_out.hasNext()) {
			i_out.next();
			QListIterator<QVPropertyHolderLink*> j_out(i_out.value());
			while(j_out.hasNext()) {
				QVPropertyHolderLink *link = j_out.next();
				link->markedForDeletion = TRUE;
				// Protect against a possible pending acquire() for our output
				// in other holders:
				link->SyncSemaphoreOut.release();
			}
		}
	}

void QVPropertyHolder::readInputProperties()
	{
	// We read every linked input property from its source, protecting
	// the read with a standard RWLock (that here we just lock for read).
	// The only caveat is that if a property is synchronously read, then
	// we must wait for the producer to write it first. We implement that
	// by waiting on the SyncSemaphoreOut of the link. Also, in this
	// case when we finish reading the property, we signal every possible
	// waiting writer that, regarding this specific link, it can write now
	// a new value if it needs to, because we have read the old value yet.
	// This is implemented by releasing the SyncSemaphoreIn associated to
	// the link.
	QMutableMapIterator<QString, QVPropertyHolderLink*> i(inputLinks);
	while (i.hasNext()) {
		i.next();
		QVPropertyHolderLink *link = i.value();
		if(link->link_type == SynchronousLink and not link->markedForDeletion) {
			link->SyncSemaphoreOut.acquire();
		}
		link->qvp_orig->RWLock.lockForRead();
		//this->setPropertyValueQVariant(link->prop_dest,link->qvp_orig->safelyCopiedVariants[link->prop_orig]);
		this->variants[link->prop_dest] = link->qvp_orig->safelyCopiedVariants[link->prop_orig];
		link->qvp_orig->RWLock.unlock();
		if(link->link_type == SynchronousLink and not link->markedForDeletion) {
			link->SyncSemaphoreIn.release();
		}
		// Possible link deletion:
		if(link->markedForDeletion) {
			i.remove();
			delete link;
		}
	}
}

void QVPropertyHolder::writeOutputProperties()
	{
	QMutableMapIterator<QString, QList<QVPropertyHolderLink*> >i(outputLinks);

	// For every QVP synchronously linked to this QVP's output, we ask
	// for permision to write a new output (that will only be possible if
	// all of these QVP's have read their inputs already):
	while (i.hasNext()) {
		i.next();
		QListIterator<QVPropertyHolderLink*> j(i.value());
		while(j.hasNext()) {
			QVPropertyHolderLink *link = j.next();
			if(link->link_type == SynchronousLink and not link->markedForDeletion) {
				link->SyncSemaphoreIn.acquire();
			}
		}
	}

	// Now, we write a new coherent state, simply protected by the
	// corresponding RWLock:
	i.toFront();
	this->RWLock.lockForWrite();
	while (i.hasNext()) {
		i.next();
		QString prop_orig = i.key();
		safelyCopiedVariants[prop_orig] = variants[prop_orig];
	}
	this->RWLock.unlock();

	// Finally, we signal to QVP's synchronously linked to this QVP's output
	// that there is a new coherent output, by unlocking our SyncLockOut
	// lock.
	i.toFront();
	while (i.hasNext()) {
		i.next();
		QMutableListIterator<QVPropertyHolderLink*> j(i.value());
		while(j.hasNext()) {
			QVPropertyHolderLink *link = j.next();
			if(link->link_type == SynchronousLink and not link->markedForDeletion) {
				link->SyncSemaphoreOut.release();
			}
			// Possible link deletion:
			if(link->markedForDeletion) {
				j.remove();
				delete link;
				if(i.value().isEmpty()) {
					i.remove();
					break;
				}
			}
		}
	}
}

void QVPropertyHolder::setLastError(QString str) const
	{ errorString = str; }

template <> bool QVPropertyHolder::parseArgument<bool>(const QString parameter, const QString value)
	{
	if (value.toLower() == "true" || value.toLower() == "false")
		{
		//variants[parameter] = QVariant::fromValue<bool>(value.toLower() == "true");
		setPropertyValue<bool>(parameter,value.toLower() == "true");
		return TRUE;
		}
	else {
		errorString = "QVPropertyHolder::parseArgument(): holder " + getName() +
					": value " + value +
					" is not a valid boolean value for parameter " +
					parameter + ".\n";
		return FALSE;
		}
	}

template <> bool QVPropertyHolder::parseArgument<int>(const QString parameter, const QString value)
	{
	bool okInt;
	int intValue = value.toInt(&okInt);
	if(not okInt)
		{
		errorString = "QVPropertyHolder::parseArgument(): holder " + getName() +
					": value " + value +
					" is not a valid integer value for parameter " +
					parameter + ".\n";
		return FALSE;
		}
	//variants[parameter] = QVariant::fromValue<int>(intValue);
	setPropertyValue<int>(parameter,intValue);
	return TRUE;
	}

template <> bool QVPropertyHolder::parseArgument<double>(const QString parameter, const QString value)
	{
	bool okDouble;
	double doubleValue = value.toDouble(&okDouble);
	if(not okDouble)
		{
		errorString = "QVPropertyHolder::parseArgument(): holder " + getName() +
					": value " + value +
					" is not a valid double value for parameter " +
					parameter + ".\n";
		return FALSE;
		}
	//variants[parameter] = QVariant::fromValue<double>(doubleValue);
	setPropertyValue<double>(parameter,doubleValue);
	return TRUE;
	}

template <> bool QVPropertyHolder::parseArgument<QString>(const QString parameter, const QString value)
	{
	//variants[parameter] = QVariant::fromValue<QString>(value);
	setPropertyValue<QString>(parameter,value);
	return TRUE;
	}
