/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvdisjointset.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <qvdta/qvdisjointset.h>
#define	MIN(X,Y)	(((X)>(Y))?(Y):(X))
#define	MAX(X,Y)	(((X)>(Y))?(X):(Y))

namespace qvdta
{
QVDisjointSet::QVDisjointSet(uInt numElements): elements(numElements), cols(0), rows(0)
	{
	Q_ASSERT_X(numElements > 0, "QVDisjointSet", "Number of elements equals 0 in constructor");
	allocData();
	makeSet();
	}

QVDisjointSet::QVDisjointSet(uInt cols, uInt rows): elements(cols*rows), cols(cols), rows(rows)
	{
	Q_ASSERT_X(cols > 0, "QVDisjointSet", "Number of columns equals 0 in constructor");
	Q_ASSERT_X(rows > 0, "QVDisjointSet", "Number of rows equals 0 in constructor");
	allocData();
	makeSet();
	}

QVDisjointSet::QVDisjointSet(QVGenericImage &image): elements(cols*rows), cols(image.getCols()), rows(image.getRows())
	{
	Q_ASSERT_X(cols > 0, "QVDisjointSet", "Number of columns equals 0 in constructor");
	Q_ASSERT_X(rows > 0, "QVDisjointSet", "Number of rows equals 0 in constructor");
	allocData();
	makeSet();
	}

QVDisjointSet::~QVDisjointSet()
	{
	Q_ASSERT_X(parent != 0, "~QVDisjointSet", "Parent array not allocated");
	Q_ASSERT_X(rank != 0, "~QVDisjointSet", "Rank array not allocated");
	freeData();
	}

uInt QVDisjointSet::unify(uInt index1, uInt index2)
	{
	Q_ASSERT_X(index1 < elements, "QVDisjointSet::unify", "First index exceeds number of elements");
	Q_ASSERT_X(index2 < elements, "QVDisjointSet::unify", "Second index exceeds number of elements");

	uInt root1 = find(index1), root2 = find(index2);

	// Trivial cases
	if ( (index1 == index2) || (root1 == root2) )
		return root1;

	// Addon to union-find algorithm to keep count of number of sets with constant cost
	sets--;

	Q_ASSERT_X(sets > 0 , "QVDisjointSet::unify", "Number of sets reached 0");
	
	if (rank[root1] > rank[root2])
		{
		uInt temp = root1;
		root1 = root2;
		root2 = temp;
		}

	Q_ASSERT_X(rank[root1] == MIN(rank[root1], rank[root2]) ,
		"QVDisjointSet::unify", "first root is minimal");
	Q_ASSERT_X(rank[root2] == MAX(rank[root1], rank[root2]) ,
		"QVDisjointSet::unify", "first root is maximal");

	if (rank[root1] == rank[root2])
		rank[root2] = rank[root2] +1;

	parent[root1] = root2;
	count[root2] += count[root1];

	Q_ASSERT_X(find(index1) == find(index2) , "QVDisjointSet::unify",
		"Parent for indexes don't coincide, after unify");

	return find(index1);
	}

void QVDisjointSet::makeSet()
	{
	sets = elements;
	for (uInt index = 0; index < elements; index++)
		{ 
		rank[index] = 0;
		count[index] = 1;
		parent[index] = index;
		}
	}

void QVDisjointSet::allocData()
	{
	parent = new uInt[elements];
	rank = new uInt[elements];
	count = new uInt[elements];
	}

void QVDisjointSet::freeData()
	{
	delete parent;
	delete rank;
	delete count;
	}

}
