/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvdta.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <qvipp/qvipp.h>
#include <qvdta/qvdta.h>

namespace qvdta
{
QVector< QVector< QPoint > > CountingSort(const QVImage<uChar, 1> &image)
	{
	QVector< QVector <QPoint> > result(256);

	QMap<sInt, int> histogram = qvipp::HistogramRange(image);
	for (int k=0; k<256; k++)
		result[k].reserve(histogram.value(k));

	QVIMAGE_INIT_READ(uChar,image);
	for(uInt row = 0; row < image.getRows(); row++)
		for(uInt col = 0; col < image.getCols(); col++)
			result[QVIMAGE_PIXEL(image, col, row,0)].append(QPoint(col, row));
	
	return result;
	}

void HarrisCornerResponseImage(const QVImage<uChar> &image, QVImage<sFloat> &result)
	{
	uInt	rows = image.getRows(), cols = image.getCols();
	QVImage<uChar> buffer;
	qvipp::MinEigenValGetBufferSize(image, buffer);
	QVImage<sFloat>	eigval(cols, rows);

	qvipp::MinEigenVal(image, eigval, buffer);

	result = eigval;
	}

void SobelCornerResponseImage(const QVImage<uChar> &image, QVImage<sFloat> &result)
	{
	uInt	cols = image.getCols(), rows = image.getRows();
	QVImage<sFloat> imageSFloat = image;

	QVImage<sFloat> Gx(cols, rows), Gy(cols, rows), Gxx(cols, rows),
			Gxy(cols, rows), Gyx(cols, rows), Gyy(cols, rows);

	qvipp::FilterSobelHorizMask(imageSFloat,Gx);
	qvipp::FilterSobelVertMask(imageSFloat,Gy);

	qvipp::FilterSobelHorizMask(Gx,Gxx);
	qvipp::FilterSobelVertMask(Gy,Gyy);

	qvipp::FilterSobelHorizMask(Gy,Gyx);
	qvipp::FilterSobelVertMask(Gx,Gxy);

	QVImage<sFloat> GxyGyx(cols, rows), GxxGyy(cols, rows);

	qvipp::Mul(Gxy, Gyx, GxyGyx);
	qvipp::Mul(Gxx, Gyy, GxxGyy);

	QVImage<sFloat> diffGxyGyx_GxxGyy(cols, rows);

	qvipp::Sub(GxyGyx, GxxGyy, diffGxyGyx_GxxGyy);

	for (uInt col = 0; col < cols; col++)
		for(uInt row = 0; row < rows; row++)
			if (diffGxyGyx_GxxGyy(col, row) <= 0)
				diffGxyGyx_GxxGyy(col, row) = 0;

	result = diffGxyGyx_GxxGyy;
	}


void FilterLocalMax(const QVImage<sFloat> &src, QVImage<uChar> &dest, uInt colMaskSize, uInt rowMaskSize, sFloat threshold)
	{
	// TODO:
	//	- Fix resulting image ROI, obtain maximums restricted to the ROI.
	//	- Use IPP functions to obtain the result.
	uInt cols = src.getCols(), rows = src.getRows();
	qvipp::Set(dest,0);
	sFloat actual;

	QVIMAGE_INIT_READ(sFloat,src);
	QVIMAGE_INIT_WRITE(uChar,dest);
	for(uInt row = rowMaskSize; row < rows-rowMaskSize; row++)
		for(uInt col = colMaskSize; col < cols-colMaskSize; col++)
			{
			actual = QVIMAGE_PIXEL(src, col, row,0);
			if (actual < threshold)
				continue;
			QVIMAGE_PIXEL(dest, col, row, 0) = IPP_MAX_8U;
			for (uInt j = row-rowMaskSize; j < row+rowMaskSize; j++)
				for (uInt i = col-colMaskSize; i < col+colMaskSize; i++)
					if (	((i != col) || (j != row)) &&
						(actual <= QVIMAGE_PIXEL(src, i, j, 0))	)
						{
						QVIMAGE_PIXEL(dest, col, row, 0) = 0;
						goto next;
						}
			next:
			continue;
			}
	}

int myFloodFill(QVImage<uChar> &image, uInt x, uInt y, uInt value, uInt minVal, uInt maxVal)
	{
	// Value should be inside range [minVal, maxVal]
	Q_ASSERT( (value <= minVal) || (value >= maxVal) );
	Q_ASSERT( minVal <= maxVal );

	// Coordinates should be inside the image.
	if ( (x >= image.getCols()) || (y >= image.getRows()))
		return 0;

	if ( (image(x,y) < minVal) || (image(x,y) > maxVal) )
		return 0;

	image(x,y) = value;

	int val = 1;
	val += myFloodFill(image, x-1, y, value, minVal, maxVal);
	val += myFloodFill(image, x, y-1, value, minVal, maxVal);
	val += myFloodFill(image, x+1, y, value, minVal, maxVal);
	val += myFloodFill(image, x, y+1, value, minVal, maxVal);

	val += myFloodFill(image, x-1, y-1, value, minVal, maxVal);
	val += myFloodFill(image, x-1, y+1, value, minVal, maxVal);
	val += myFloodFill(image, x+1, y-1, value, minVal, maxVal);
	val += myFloodFill(image, x+1, y+1, value, minVal, maxVal);

	return val;
	}

#define DEFINE_QVDTA_FUNCTION_EQUALIZE_HISTOGRAM(TYPE, C)				\
void equalizeHistogram(const QVImage<TYPE,C> &image, QVImage<TYPE,C> &equalized)	\
	{										\
	uInt	rows = image.getRows(), cols = image.getCols();				\
	TYPE	maxVal, minVal;								\
											\
	qvipp::Max(image,maxVal);							\
	qvipp::Min(image,minVal);							\
											\
	QVImage<TYPE,C> temp(cols, rows), result(cols, rows);				\
	qvipp::SubC(image, temp, minVal);						\
	qvipp::MulC(temp, result, 255/(maxVal-minVal));					\
	equalized = result;								\
	}

DEFINE_QVDTA_FUNCTION_EQUALIZE_HISTOGRAM(uChar,1);
DEFINE_QVDTA_FUNCTION_EQUALIZE_HISTOGRAM(sFloat,1);
}
