/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvcanvas.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVCANVAS_H
#define QVCANVAS_H

#include <QGLWidget>
#include <QPainter>
#include <QString>

#include <qvcore/qvimage.h>

class QwtScaleWidget;
class QwtLinearScaleEngine;
class QwtScaleDiv;
class QToolButton;
class QStatusBar;
class QVImageArea;

/// @class QVPainter qvcanvas.h qvgui/qvcanvas.h
/// @brief Derived class that adds QVImage's drawing capabilities to a QPainter.
///
/// This class adds any kind of QVImage (subclasses of QVGenericImage) drawing
/// capabilities to a QPainter, when drawing in a QVCanvas. The final 
/// library user should only use it as returned by the
/// QVCanvas::getQVPainter() method.
/// @ingroup qvgui
class QVPainter: public QPainter
{
friend class QVImageArea;

private:
	// Nobody except QVImageArea should create or destroy QVPainters:
	QVPainter(QVImageArea *imageArea): QPainter()
		{ this->imageArea = imageArea;};
	~QVPainter() { };
public:
	void drawQVImage(QVGenericImage *image,bool adaptsize=TRUE,float low=0.0, float high=255.0);
	void drawTextUnscaled(const QPointF & position, const QString & text);
	void drawTextUnscaled(const QPoint & position, const QString & text);
	void drawTextUnscaled(const QRectF & rectangle, int flags,
						  const QString & text, QRectF * boundingRect = 0);
	void drawTextUnscaled(const QRect & rectangle, int flags,
						  const QString & text, QRect * boundingRect = 0);
	void drawTextUnscaled(int x, int y, const QString & text);
	void drawTextUnscaled(int x, int y, int width, int height, int flags,
						  const QString & text, QRect * boundingRect = 0);
	void drawTextUnscaled(const QRectF & rectangle, const QString & text,
						  const QTextOption & option = QTextOption());

private:
	// QVPainter keeps a private reference to its corresponding QVImageArea:
	QVImageArea *imageArea; 
};


/************************* QVImageArea auxiliary class ************************/

// Auxiliary class, not available to library users (only to be used by
// QVCanvas, which will be its friend). Thus, everything is private.
// Note.- Not declared as a nested class because of QT moc limitations
// with the Q_OBJECT macro.

/// \cond 	Doxygen will ignore from here...
class QVImageArea : public QGLWidget
{
	Q_OBJECT

friend class QVCanvas;
friend class QVPainter;

private:
	// static class pointer to first instance of QVImageArea. The rest of
	// imageAreas will share context with it, in order to be fast enough
	// (otherwise, use of multiple context can make the application very
	// slow on some graphic cards!).
	static QVImageArea *first_image_area;
	
	void initObject(int w, int h); // Auxiliary function for constructors.

	// Nobody except QVCanvas should create or destroy QVImageArea's:
	// Constructor for first instance:
	QVImageArea(int w, int h,QWidget *parent);
	// Constructor for second and subsequent instances:
	QVImageArea(int w, int h,QWidget *parent,QGLWidget *other);
	~QVImageArea() {};
	
	enum TMouseMode {
		noneMode = 0x01,
		dragMode = 0x02,
		selMode = 0x03,
		zoomMode = 0x04
	};

signals:
	void newGeometry(int origheight,int origwidth,int topleftx,int toplefty,int width,int height, int zoom);
	void newMousePosition(float x,float y);
	void mouseLeavesImageArea(bool leaves);

protected:
	// void initializeGL() {};
	void wheelEvent(QWheelEvent *event);
	void resizeGL(int width, int height);
	void paintEvent(QPaintEvent *event);
	void mousePressEvent(QMouseEvent *event);
	void mouseMoveEvent(QMouseEvent *event);
	void mouseReleaseEvent(QMouseEvent *event);
	void leaveEvent(QEvent *event);

private:
	const int max_zoom;
	void drawQVImage(QVGenericImage *image,bool adaptsize,float low,float high);
	void centerZoom(int zoom);
	void resizeImageArea(int w,int h);
	int zoom,origheight,origwidth;
	QPoint topLeft;
	QRect selRect,zoomRect;
	QRect innerRect();
	QRect outerRect();
	TMouseMode mouseMode;
	QPoint firstPos,lastPos;
	bool dragging;
	QRectF intuitiveRect(QRect rect);
	QVPainter *painter;
	QList<QVGenericImage*> imageList;
};

/// \endcond	...to here

class QVCanvas : public QWidget
{
/// \cond
friend class QVImageArea;
/// \endcond

	Q_OBJECT
public:
	QVCanvas(int w = 200,int h = 100,QWidget *parent = 0);
	~QVCanvas();
	int getZoom() { return imageArea->zoom; }
	QRect getViewport() { return QRect(imageArea->topLeft,QSize(imageArea->width(),imageArea->height())); } //OJO, document in the manual that all viewport is multiplyed by actual zoom.
	QSize getSize() { return QSize(imageArea->origwidth,imageArea->origheight); }
	QVPainter *getQVPainter() { return imageArea->painter; };
	QRect getSelectionRectangle() { return imageArea->selRect; }

protected:
	virtual void viewer() { /* Default is to paint nothing. */ };

signals:
	void newGeometry(int origwidth,int origheight,int topleftx,int toplefty,int width,int height, int zoom);
	//void refreshed();

public slots:
	void setGeometry(int origwidth,int origheight,int topleftx,int toplefty,int width,int height, int zoom);
	void refreshImageArea();

private slots:
	void zoomRectClicked(bool checked);
	void selRectClicked(bool checked);
	void dragClicked(bool checked);
	void zoomInClicked();
	void zoomOutClicked();
	void zoomOriginalClicked();
	void newMousePositionSlot(float x,float y);
	void mouseLeavesImageAreaSlot(bool leaves);

private:
	QwtScaleWidget *scaleWidgetX,*scaleWidgetY;
	QwtLinearScaleEngine *scaleEngineX,*scaleEngineY;
	QVImageArea *imageArea;
	QToolButton *buttonZoomIn,*buttonZoomOut,*buttonZoomOriginal,*buttonZoomRect,*buttonSelRect,*buttonDrag;
	QStatusBar *statusBar;
	void resizeEvent(QResizeEvent *event);
	int scaleWidgetsFixedWidth,statusBarWidgetFixedHeight;
	float mousePosX,mousePosY;
	bool mouseIsOut;
	QString statusMessage();
};

#endif
